from typing import Optional, Mapping
from dateutil import parser
from datetime import datetime

import re
from aiohttp import ClientResponse, ClientSession
from .types import (Word, Meaning,
                    BriefMeaning, Translation,
                    PartOfSpeechCode, Definition,
                    Properties, Example,
                    MeaningWithSimilarTranslation,
                    AlternativeTranslation, Pronunciation, Language)


class Dictionary:
    def __init__(self) -> None:
        self.url_search = f"https://dictionary.skyeng.ru/api/public/v1/words/search"
        self.url_meaning = f"https://dictionary.skyeng.ru/api/public/v1/meanings"

    async def __aenter__(self):
        self._session = ClientSession(raise_for_status=True)
        return self

    async def __aexit__(self, *args, **kwargs):
        await self._close()

    async def _fetch(
            self,
            url: str,
            *,
            params: Optional[Mapping[str, str]],
            headers: Mapping[str, str],
            session: ClientSession,
    ) -> ClientResponse:
        async with session.get(url, params=params, headers=headers, raise_for_status=True) as response:
            await response.read()
        return response

    def _what_part_of_speech(self, part: str) -> PartOfSpeechCode:
        return PartOfSpeechCode.__dict__['_member_map_'][part]

    def _get_words(self, data) -> list[Word]:
        words = []
        if data is not None:
            for word in data:
                brief_meanings = []
                for brief_meaning in word['meanings']:
                    part_of_speech_code = self._what_part_of_speech(brief_meaning['partOfSpeechCode'])
                    match = re.search(pattern=r"https.+", string=brief_meaning['imageUrl'])
                    m = BriefMeaning(
                        id=int(brief_meaning['id']),
                        part_of_speech_code=part_of_speech_code,
                        translation=brief_meaning['translation']['text'],
                        translation_note=brief_meaning['translation']['note'],
                        image_url=None if match is None else match.group(0),
                        transcription=brief_meaning['transcription'],
                        sound_url=Pronunciation(brief_meaning['soundUrl'], Language.en)  # sound of transcription
                    )
                    brief_meanings.append(m)
                word = Word(
                    id=int(word['id']),
                    text=word['text'],
                    meanings=brief_meanings
                )
                words.append(word)
        return words

    def _get_meanings(self, data) -> list[Meaning]:
        meanings = []
        if data is not None:
            for meaning in data:
                images = []
                for image in meaning['images']:
                    match = re.search(pattern=r"https.+", string=image['url'])
                    images.append(None if match is None else match.group(0))
                examples = []
                for example in meaning['examples']:
                    examples.append(Example(text=example['text'],
                                            sound_url=Pronunciation(example['soundUrl'], Language.en)))
                meanings_with_similar_translation = []
                for meaning_with_similar_translation in meaning['meaningsWithSimilarTranslation']:
                    form = MeaningWithSimilarTranslation(
                        meaning_id=meaning_with_similar_translation['meaningId'],
                        frequency_percent=meaning_with_similar_translation['frequencyPercent'],
                        part_of_speech_abbreviation=meaning_with_similar_translation['partOfSpeechAbbreviation'],
                        translation=meaning_with_similar_translation['translation']['text'],
                        translation_note=meaning_with_similar_translation['translation']['note']
                    )
                    meanings_with_similar_translation.append(form)
                alternative_translations = []
                for alternative_translation in meaning['alternativeTranslations']:
                    form = AlternativeTranslation(
                        text=alternative_translation['text'],
                        translation=alternative_translation['translation']['text'],
                        translation_note=alternative_translation['translation']['note']
                    )
                    alternative_translations.append(form)
                _prop = meaning['properties']
                properties = Properties(collocation=_prop.get('collocation'),
                                        not_gradable=_prop.get('notGradable'),
                                        irregular=_prop.get('irregular'),
                                        past_tense=_prop.get('pastTense'),
                                        past_participle=_prop.get('pastTense'),
                                        countability=_prop.get('countability'),
                                        plurality=_prop.get('plurality'),
                                        plural=_prop.get('plural'),
                                        irregular_plural=_prop.get('irregularPlural'),
                                        transitivity=_prop.get('transitivity'),
                                        linking_verb=_prop.get('linkingVerb'),
                                        linking_type=_prop.get('linkingType'),
                                        phrasal_verb=_prop.get('phrasalVerb'),
                                        sound_url=_prop.get('soundUrl'),
                                        false_friends=_prop.get('falseFriends'), )
                part_of_speech_code = self._what_part_of_speech(meaning['partOfSpeechCode'])
                meaning = Meaning(
                    id=meaning['id'],
                    word_id=int(meaning['wordId']),
                    difficulty_level=meaning['difficultyLevel'],
                    part_of_speech_code=part_of_speech_code,
                    prefix=meaning['prefix'],
                    text=meaning['text'],
                    sound_url=Pronunciation(meaning['soundUrl'], Language.en),
                    transcription=meaning['transcription'],
                    properties=properties,
                    updated_at=meaning['updatedAt'],
                    mnemonics=meaning['mnemonics'],
                    translation=meaning['translation']['text'],
                    translation_note=meaning['translation']['note'],
                    images=[None] if images == [] else images,
                    definition=meaning['definition']['text'],
                    definition_sound_url=meaning['definition']['soundUrl'],
                    examples=examples,
                    meanings_with_similar_translation=meanings_with_similar_translation,
                    alternative_translations=alternative_translations
                )
                meanings.append(meaning)
        return meanings

    async def words(self, word: str, page: int = 1, pagesize: int = 0) -> list[Word]:
        '''
        :param word:
        :param page:
        :param pagesize: if value 0 then show full page, if value 1 then show only one word
        :return: Find words with meanings. It will return relevant meanings grouped by words.
                 You can search in English or translation
        '''
        params = {
            'search': word,
            'page': page,
            'pageSize': pagesize}
        headers = {}
        response = await self._fetch(self.url_search, params=params, headers=headers, session=self._session)
        # print(response.status)
        data = await response.json()
        return self._get_words(data)

    def _out_value(self, ids: int | list[int]) -> str:
        '''
        :param ids: id or array ids
        :return: A string of meaning ids. Separated by a comma
        '''
        if type(ids) is int:
            return str(ids)
        elif type(ids) is list:
            return ','.join(str(el) for el in ids)
        else:
            raise ValueError

    def _convert_time(self, complex_date_string: str) -> str:
        default = datetime(2000, 1, 1, 0, 0, 0)
        date_obj_flex = parser.parse(complex_date_string, default=default)
        return date_obj_flex.strftime("%Y-%m-%d %H:%M:%S")

    async def meaning(self, ids: int | list[int], data: str = None) -> list[Meaning]:
        '''
        :param ids: An array of meaning ids or integer value
        :param data: Retrieve results from this date. * * Must be in UTC timezone. *
        :return: Return full information about meaning
        '''
        params = {
            'ids': self._out_value(ids),
            'updatedAt': '' if data is None else self._convert_time(data)
        }
        headers = {}
        print(params)
        response = await self._fetch(self.url_meaning, params=params, headers=headers, session=self._session)
        data = await response.json()
        return self._get_meanings(data)

    async def _close(self) -> None:
        if not self._session.closed:
            # print("Session is closed")
            await self._session.close()



from dataclasses import dataclass
from typing import TypeAlias
from enum import Enum
from urllib.parse import urlparse, parse_qs, urlencode

MeaningId: TypeAlias = int

class PartOfSpeechCode(Enum):  # Available code
    def __init__(self, en, ru):
        self.en = en
        self.ru = ru

    n = ("noun", "существительное")
    v = ("verb", "глагол")
    j = ("adjective", "прилагательное")
    r = ("adverb", "наречие")
    prp = ("preposition", "предлог")
    prn = ("pronoun", "местоимение")
    crd = ("cardinal number", "количественное числительное")
    crj = ("conjunction", "союз")
    exc = ("interjection", "междометие")
    det = ("article", "артикль")
    abb = ("abbreviation", "сокращение")
    x = ("particle", "частица")
    ord = ("ordinal number", "порядковое числительное")
    md = ("modal verb", "модальный глагол")
    ph = ("phrase", "фраза")
    phi = ("idiom", "идиома")


@dataclass
class Translation:
    text: str  # A text of a translation.
    note: str  # A note about translation.


class Language(Enum):
    en = 'english'
    ru = 'russian'

class Pronunciation:
    '''
    handler link with own params
    example: https://vimbox-tts.skyeng.ru/api/v1/tts?text=Lacking+ease+or+grace.&lang=en&voice=male_2'
    '''
    def __init__(self, url: str, language: Language):
        __parse_result = urlparse(url)
        self.__scheme = __parse_result.scheme
        self.__netloc = __parse_result.netloc
        self.__path = __parse_result.path
        __params = parse_qs(__parse_result.query)
        __text = __params.get('text')[0]
        __lang = language.name
        self.__assembly_params_dict = {'text': __text,
                                       'lang': __lang,
                                       'voice': 'male_1'}
        if __params.get('isSsml') is not None:
            __is_ssml = __params.get('isSsml')[0]
            self.__assembly_params_dict['isSsml'] = __is_ssml

    @property
    def male_1(self):
        self.__assembly_params_dict['voice'] = 'male_1'
        params = urlencode(self.__assembly_params_dict)
        return f'{self.__scheme}://{self.__netloc}{self.__path}?{params}'

    @property
    def male_2(self):
        self.__assembly_params_dict['voice'] = 'male_2'
        params = urlencode(self.__assembly_params_dict)
        return f'{self.__scheme}://{self.__netloc}{self.__path}?{params}'

    @property
    def female_1(self):
        self.__assembly_params_dict['voice'] = 'female_1'
        params = urlencode(self.__assembly_params_dict)
        return f'{self.__scheme}://{self.__netloc}{self.__path}?{params}'

    @property
    def female_2(self):
        self.__assembly_params_dict['voice'] = 'female_2'
        params = urlencode(self.__assembly_params_dict)
        return f'{self.__scheme}://{self.__netloc}{self.__path}?{params}'

    def __repr__(self):
        return self.male_1

    def __str__(self):
        return self.male_1


@dataclass
class BriefMeaning:  # Meaning2
    id: MeaningId  # MeaningId.
    part_of_speech_code: PartOfSpeechCode
    translation: str
    translation_note: str
    image_url: str  #
    transcription: str  #
    sound_url: Pronunciation


@dataclass
class Properties:
    collocation: bool | None #
    not_gradable: bool | None
    irregular: bool | None #
    past_tense: str | None  # "found",
    past_participle: str | None  # "found",
    transitivity: str | None  # "t", "i" "ti"
    countability: str | None # "c", cu
    plurality: str | None # s , sp
    plural: str | None
    irregular_plural: str | None
    phrasal_verb: bool | None # false,
    linking_verb: bool | None #
    linking_type: str | None # 'L + noun, L + adjective'
    sound_url: Pronunciation | None  #
    false_friends: list | None   #


@dataclass  #
class Definition:  #
    text: str  #
    sound_url: Pronunciation  #


@dataclass
class Example:
    text: str  #
    sound_url: Pronunciation  #


@dataclass
class MeaningWithSimilarTranslation:
    meaning_id: int  #
    frequency_percent: float  #
    part_of_speech_abbreviation: str  # часть речи на русском напрм: "гл."
    translation: str  #
    translation_note: str


@dataclass
class AlternativeTranslation:  #
    text: str  #
    translation: str  #
    translation_note: str


@dataclass
class Meaning:
    id: MeaningId  # Meaning id.
    word_id: int  # Word is a group of meanings. We combine meanings by word entity.
    difficulty_level: int  # There are 6 difficultyLevels: 1, 2, 3, 4, 5, 6.
    part_of_speech_code: PartOfSpeechCode  # String representation of a part of speech.
    prefix: str  # Infinitive particle (to) or articles (a, the).
    text: str  # Meaning text.
    sound_url: Pronunciation  # URL to meaning sound.
    transcription: str  # IPA phonetic transcription.
    properties: Properties  #
    updated_at: str  #
    mnemonics: str  #
    translation: str  #
    translation_note: str | None
    images: list[str]  # A collection of an images.
    definition: str  #
    definition_sound_url: Pronunciation
    examples: list[Example]  # Usage examples
    meanings_with_similar_translation: (
        list[MeaningWithSimilarTranslation])  # Collection of meanings with similar translations.
    alternative_translations: (
        list[AlternativeTranslation])  # Collection of alternative translations.


@dataclass
class Word:
    id: int  #
    text: str  #
    meanings: list[BriefMeaning]  #
